/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLER_CPP
#define ASSEMBLER_CPP

#include "Assembler.h"


#include "Read.h"
#include "ScoredSeqSubseq.h"
#include "ScoredSeqNested.h"
#include "ScoredSeqMonoScore.h"
#include "OutputFileNull.h"
#include "OutputFileFasta.h"
#include "OutputFilePriceq.h"
#include "OutputFileMulti.h"
#include "AssemblerListenerNull.h"
#include "AssemblerListenerStd.h"
#include "AssemblerListenerVerbose.h"
#include "AssemblerListenerCarrier.h"
#include <sstream>
#include <omp.h>

ParamsMinOverlap* Assembler::getDefaultParamsMinOverlap(){ return new ParamsMinOverlap(35,20,5); }


Assembler::Assembler(){}
Assembler::Assembler(ParamsCycleManagement* cycleInfo) {
  _cycleInfo = new ParamsCycleManagement(cycleInfo);
  constructorHelper();
}
Assembler::Assembler(int totalCycles) {
  _cycleInfo = new ParamsCycleManagement(totalCycles);
  constructorHelper();
}
Assembler::Assembler(int totalCycles, long maxContigLinkages) {
  _cycleInfo = new ParamsCycleManagement(totalCycles,maxContigLinkages);
  constructorHelper();
}
void Assembler::constructorHelper(){
  _maxThreadsPerFile = 1;
  ParamsMinOverlap* defPmo = getDefaultParamsMinOverlap();
  ParamsMinFractId* defPmf = ParamsMinFractId::getDefault();
  AlignmentScoreMatrix* defAsm = AlignmentScoreMatrix::getDefault();
  _miniAlignParams = new ParamsAlignment(defPmo, defPmf, defAsm);
  _metaAlignParams = new ParamsAlignment(defPmo, defPmf, defAsm);
  _paramsDeBruijn = new ParamsDeBruijn();
  delete defPmo;
  delete defPmf;
  delete defAsm;
  _stdOutListener = new AssemblerListenerStd();
  _fileListener = NULL;
  EcoFilter* defaultEcoFilterMinCount = EcoFilterMinCount::defaultFilter();
  _cycleInfo->addEcoFilter(defaultEcoFilterMinCount);
  delete defaultEcoFilterMinCount;
  _waitingTargetMode = NULL;
}

void Assembler::setMiniAlignParams(ParamsAlignment* params){
  delete _miniAlignParams;
  _miniAlignParams = new ParamsAlignment(params);
}
void Assembler::setMetaAlignParams(ParamsAlignment* params){
  delete _metaAlignParams;
  _metaAlignParams = new ParamsAlignment(params);
}



Assembler::~Assembler(){
  for (set<ScoredSeq*>::iterator it = _currentCycleContigs.begin(); it != _currentCycleContigs.end(); ++it){ (*it)->deepDelete(); }
  delete _miniAlignParams;
  delete _metaAlignParams;
  delete _cycleInfo;
  delete _paramsDeBruijn;
  for (set<ParameterizedReadFile*>::iterator it = _inputPairedReadFiles.begin(); it != _inputPairedReadFiles.end(); ++it){ delete (*it); }
  for (set<OutputFile*>::iterator it = _outfileSet.begin(); it != _outfileSet.end(); ++it){ delete *it; }
  delete _stdOutListener;
  delete _fileListener;
  for (vector<ReadPairFilter*>::iterator it = _waitingRpf.begin(); it != _waitingRpf.end(); ++it){ delete *it; }
  for (set<ParameterizedInitialFile*>::iterator it = _inputInitialContigFiles.begin(); it != _inputInitialContigFiles.end(); ++it){
    delete *it;
  }
  delete _waitingTargetMode;
}



void Assembler::runSetup(){

  // initial contig files
  _initialContigFiles.insert( _inputInitialContigFiles.begin(), _inputInitialContigFiles.end() );
  _initialContigFiles.insert( _inputInitialContigFilesNoTarget.begin(), _inputInitialContigFilesNoTarget.end() );
  // and set up the target mode
  if (_waitingTargetMode != NULL){
    _waitingTargetMode->addFiles( &_inputInitialContigFiles );
    _cycleInfo->addEcoFilter(_waitingTargetMode);
  }

  // read files
  ParameterizedReadFile* rfArray[ _inputPairedReadFiles.size() + 1 ];
  int prfCount = 0;
  for (set<ParameterizedReadFile*>::iterator it = _inputPairedReadFiles.begin();
       it != _inputPairedReadFiles.end(); ++it){
    rfArray[prfCount] = *it;
    prfCount++;
  }
  long rfSizes[ prfCount + 1 ];

  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < prfCount; ++n){ rfSizes[n] = rfArray[n]->numReads(); }

  int numThreads = omp_get_max_threads();
  long readTotal = 0;
  for (int n = 0; n < prfCount; ++n){ readTotal += rfSizes[n]; }
  long threadDenom = readTotal / numThreads;
  if (threadDenom == 0){ threadDenom = 1; }

  #pragma omp parallel for schedule(dynamic)
  for (int n = 0; n < prfCount; ++n){
    int rfThreads = rfSizes[n] / threadDenom;
    if (rfThreads == 0){ rfThreads = 1; }
    else if (rfThreads > _maxThreadsPerFile){ rfThreads = _maxThreadsPerFile; }
    set<ParameterizedReadFile*> splitFiles;
    rfArray[n]->splitFile(&splitFiles, rfThreads);
    #pragma omp critical (Assembler_addRfsToSet)
    { _pairedReadFiles.insert( splitFiles.begin(), splitFiles.end() ); }
  }

  // output file(s)
  if (_outfileSet.size() == 0){ _outfile = new OutputFileNull(); }
  else {
    OutputFile** outfileArray = new OutputFile*[ _outfileSet.size() ];
    int arrayIndex = 0;
    for (set<OutputFile*>::iterator it = _outfileSet.begin(); it != _outfileSet.end(); ++it){
      outfileArray[arrayIndex] = *it;
      arrayIndex++;
    }
    _outfile = new OutputFileMulti(outfileArray, arrayIndex);
    delete [] outfileArray;
  }

  // the read pair filters
  if (_waitingRpf.size()==0){ _rpf = new ReadPairFilterNull(); }
  else if (_waitingRpf.size()==1){ _rpf = *(_waitingRpf.begin()); }
  else {
    ReadPairFilter** filters = new ReadPairFilter*[ _waitingRpf.size() ];
    int rpfIndex = 0;
    for (vector<ReadPairFilter*>::iterator rpfIt = _waitingRpf.begin(); rpfIt != _waitingRpf.end(); ++rpfIt){
      filters[rpfIndex] = *rpfIt;
      rpfIndex++;
    }
    _rpf = new ReadPairFilterMulti(filters,rpfIndex);
    delete [] filters;
  }

  // the listener
  if (_fileListener == NULL){ _listener = _stdOutListener; }
  else {
    AssemblerListener* listenerArray[] = {_stdOutListener, _fileListener};
    _listener = new AssemblerListenerCarrier(listenerArray, 2);
  }
}


void Assembler::runCleanup(){
  delete _outfile;
  if (_fileListener != NULL){ delete _listener; }
  for (set<ParameterizedReadFile*>::iterator it = _pairedReadFiles.begin(); it != _pairedReadFiles.end(); ++it){
    if (_inputPairedReadFiles.count(*it) == 0){ delete *it; }
  }
  _pairedReadFiles.clear();
  _initialContigFiles.clear();
  if (_waitingRpf.size() != 1){ delete _rpf; }
}


void Assembler::addInitialFile(string filename, int numSteps, int cyclesPerStep, float countFactor){
  ReadFile* file = ReadFile::makeBasicReadFile(filename, countFactor);
  ParameterizedInitialFile* pif = new ParameterizedInitialFile( file, numSteps, cyclesPerStep );
  delete file;
  _inputInitialContigFiles.insert(pif);
}
void Assembler::addInitialFile(long numSeqsToUse, string filename, int numSteps, int cyclesPerStep, float countFactor){
  ReadFile* file = ReadFile::makeBasicReadFile(filename, countFactor);
  ParameterizedInitialFile* pif = new ParameterizedInitialFile( numSeqsToUse, file, numSteps, cyclesPerStep );
  delete file;
  _inputInitialContigFiles.insert(pif);
}
void Assembler::addInitialFileNoTarget(string filename, int numSteps, int cyclesPerStep, float countFactor){
  ReadFile* file = ReadFile::makeBasicReadFile(filename, countFactor);
  ParameterizedInitialFile* pif = new ParameterizedInitialFile( file, numSteps, cyclesPerStep );
  delete file;
  _inputInitialContigFilesNoTarget.insert(pif);
}
void Assembler::addInitialFileNoTarget(long numSeqsToUse, string filename, int numSteps, int cyclesPerStep, float countFactor){
  ReadFile* file = ReadFile::makeBasicReadFile(filename, countFactor);
  ParameterizedInitialFile* pif = new ParameterizedInitialFile( numSeqsToUse, file, numSteps, cyclesPerStep );
  delete file;
  _inputInitialContigFilesNoTarget.insert(pif);
}



void Assembler::addReadFile(string filename, long ampSize){
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFile(string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int numCycles, string filename, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int numCycles, string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int onCycles, int skipCycles, string filename, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int onCycles, int skipCycles, string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}


void Assembler::addReadFile(string filenameA, string filenameB, long ampSize){
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFile(string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int onCycles, int skipCycles, string filenameA, string filenameB, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addReadFileLimitUse(int startCycle, int onCycles, int skipCycles, string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makePairedReadFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}


// MATE PAIRS BEGIN

void Assembler::addMatePairFile(string filename, long ampSize){
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFile(string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int numCycles, string filename, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int numCycles, string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int onCycles, int skipCycles, string filename, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int onCycles, int skipCycles, string filename, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filename, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}


void Assembler::addMatePairFile(string filenameA, string filenameB, long ampSize){
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFile(string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int numCycles, string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, numCycles) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int onCycles, int skipCycles, string filenameA, string filenameB, long ampSize){
  ParamsMapping* mapParams = ParamsMapping::getDefault();
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles ) );
  delete rf;
  delete mapParams;
}
void Assembler::addMatePairFileLimitUse(int startCycle, int onCycles, int skipCycles, string filenameA, string filenameB, long ampSize, float fractId){
  if (fractId < .25 or fractId > 1){ throw AssemblyException::ArgError("% ID required for mapping is out of range 25-100"); }
  ParamsMapping* defMapParams = ParamsMapping::getDefault();
  ParamsMapping* mapParams = new ParamsMapping(defMapParams->stepSize(), fractId );
  delete defMapParams;
  ReadFile* rf = ReadFile::makeMatePairFile(filenameA, filenameB, ampSize);
  _inputPairedReadFiles.insert( new ParameterizedReadFile( rf, mapParams, startCycle, onCycles, skipCycles) );
  delete rf;
  delete mapParams;
}

// MATE PAIRS END


void Assembler::addOutputFile(string filename){
  _outfileSet.insert( OutputFile::makeOutputFile(filename) );
}


void Assembler::addBadSequenceFilter(string badSeqFile, float fractId){
  ReadFile* rf = ReadFile::makeBasicReadFile(badSeqFile);
  set<ScoredSeq*> badSeqs;
  rf->open();
  while ( rf->hasRead() ){ badSeqs.insert( rf->getRead() ); }
  rf->close();
  ReadPairFilter* rpf = new ReadPairFilterAvoidSeqs(&badSeqs, fractId);
  _waitingRpf.push_back( new ReadPairFilterAvoidSeqs(&badSeqs, fractId) );
  for (set<ScoredSeq*>::iterator badIt = badSeqs.begin(); badIt != badSeqs.end(); ++badIt){ delete *badIt; }
  delete rf;
}
void Assembler::addHomopolymerFilter(long maxLength){
  _waitingRpf.push_back( new ReadPairFilterHomoPolymer(maxLength) );
}



void Assembler::addLengthFilter(long minLength, int cyclesToSkip){
  EcoFilter* ecoFilterLength = new EcoFilterMinLength(minLength,cyclesToSkip);
  _cycleInfo->setSizeFilter(ecoFilterLength);
  delete ecoFilterLength;
}
void Assembler::addTargetMode(float minFractId, int cyclesToSkip, bool fullFile){
  delete _waitingTargetMode;
  _waitingTargetMode = new EcoFilterInitialContigMatch(minFractId,cyclesToSkip,fullFile);
}
void Assembler::addTargetMode(float minFractId, int cyclesToSkip,
			      int numFilterCycles, int numSkipCycles, bool fullFile){
  delete _waitingTargetMode;
  _waitingTargetMode = new EcoFilterInitialContigMatch(minFractId,cyclesToSkip,numFilterCycles,numSkipCycles,fullFile);
}



void Assembler::verboseLogFile(string filename){
  delete _fileListener;
  _fileListener = new AssemblerListenerVerbose( filename );
}
void Assembler::makeLogNull(){
  delete _stdOutListener;
  _stdOutListener = new AssemblerListenerNull();
}
void Assembler::makeLogVerbose(){
  delete _stdOutListener;
  _stdOutListener = new AssemblerListenerVerbose();
}


void Assembler::setMaxThreadsPerFile(int maxThreadsPerFile){
  _maxThreadsPerFile = maxThreadsPerFile;
}

void Assembler::setDeBruijnKmerSize(int kmerSize){
  _paramsDeBruijn->setKmerSize(kmerSize);
}
void Assembler::setDeBruijnMaxSeqLength(long maxSeqLength){
  _paramsDeBruijn->setMaxSeqLength(maxSeqLength);
}
void Assembler::setDeBruijnMinSeqNum(long minSeqNum){
  _paramsDeBruijn->setMinNumSeqs(minSeqNum);
}



void Assembler::runAssembly(){
  runSetup();
  _listener->beginAssembly("BEGIN ASSEMBLY");
  for (int n = 0; n < _cycleInfo->totalCycles(); ++n){
    _listener->beginCycle(n+1);
    generateInitialContigs(n);
    extendOneCycle(n);

    // make the number into an addendum string and write the file
    if ( _cycleInfo->outputThisCycle(n+1) ){
      stringstream addendum;
      addendum << "cycle" << n+1;
      writeOutfile( addendum.str() );
    }
    _listener->endCycle(&_currentCycleContigs);
  }
  // delete the contigs from the prior cycle in the end
  for (set<ScoredSeq*>::iterator it = _priorCycleContigs.begin(); it != _priorCycleContigs.end(); ++it){
    (*it)->deepDelete();
  }
  _listener->endAssembly("END ASSEMBLY");
  runCleanup();
}

void Assembler::writeOutfile(string addendum){
  if (_outfile->getName(addendum) != ""){
    _listener->writeOutfile( _outfile->getName(addendum) );
  }
  _outfile->open(addendum);
  ScoredSeq** sortedContigs = OutputFile::makeSortedContigArray( &_currentCycleContigs );
  _outfile->writeContigs(sortedContigs, &_currentCycleLeftovers);
  delete [] sortedContigs;
  _outfile->close();
}



void Assembler::getContigs(set<ScoredSeq*>* outputContigs){
  for (set<ScoredSeq*>::iterator it = _currentCycleContigs.begin(); it != _currentCycleContigs.end(); ++it){
    outputContigs->insert( (*it)->shallowCopy() );
  }
}



void Assembler::findLeftoverContigs(set<ScoredSeq*>* current, set<ScoredSeq*>* prior, set<ScoredSeq*>* leftover){
  // collect the sequences of the contigs being input into this assembly round
  set<string> seqsFromPrior;
  for (set<ScoredSeq*>::iterator it = prior->begin(); it != prior->end(); ++it){
    char* tempSeq = (*it)->getSeq('+');
    seqsFromPrior.insert( string(tempSeq) );
    delete [] tempSeq;
  }
  // now identify the contigs from the current cycle that didn't make progress from the last
  for (set<ScoredSeq*>::iterator it = current->begin(); it != current->end(); ++it){
    char* tempPlus = (*it)->getSeq('+');
    char* tempMinus =  (*it)->getSeq('-');
    if (seqsFromPrior.count( tempPlus ) > 0 or 
	seqsFromPrior.count( tempMinus ) > 0 ){
      leftover->insert( (*it) );
    }
    delete [] tempPlus;
    delete [] tempMinus;
  }
}


void Assembler::generateInitialContigs(int cycleNum){
  set<ScoredSeq*> initialSeqs;

  // put the current contigs in with the new initial seqs
  initialSeqs.insert( _currentCycleContigs.begin(), _currentCycleContigs.end() );
  _currentCycleContigs.clear();

  long initialCount = initialSeqs.size();
  for (set<ParameterizedInitialFile*>::iterator rfIt = _initialContigFiles.begin(); rfIt != _initialContigFiles.end(); ++rfIt){
    (*rfIt)->getContigs(&initialSeqs, cycleNum);
  }

  // TRIM
  _cycleInfo->executeTrim( &initialSeqs, -1 );

  if ( initialCount == initialSeqs.size() ){
    _currentCycleContigs.insert( initialSeqs.begin(), initialSeqs.end() );
  } else {
    _listener->beginAddInitialContigs(&initialSeqs);

    // condense (assemble) the initial contigs (assembly is double-stranded)
    AssemblyJobFactory * contigFactory = new AssemblyJobFactory(_metaAlignParams,_paramsDeBruijn,AssemblyJob::doubleStranded,_listener);
    AssemblyJob * ajr2 = contigFactory->redundancyJob( &initialSeqs );
    delete contigFactory;
    ajr2->runJob(AssemblyJob::threaded);
    // add the output of the initial assembly to the collection
    // currentCycleContigs is the initial input; priorCycleContigs is initially empty
    ajr2->allAssembledSeqs( &_currentCycleContigs );

    set<ScoredSeq*> usedSeqs;
    ajr2->discardedInputSeqs( &usedSeqs );
    for (set<ScoredSeq*>::iterator it = usedSeqs.begin(); it != usedSeqs.end(); ++it){ (*it)->deepDelete(); }
    delete ajr2;

    _listener->endAddInitialContigs(&_currentCycleContigs);
  }
}


void Assembler::extendOneCycle(int cycleNum){

  // identify the contigs that have not changed since the previous cycle (unless this is a "reset" cycle)
  set<ScoredSeq*> leftoverContigs;
  if ( ! _cycleInfo->shouldContigsReset(cycleNum+1) ){
    findLeftoverContigs(&_currentCycleContigs, &_priorCycleContigs, &leftoverContigs);
  }
  for (set<ScoredSeq*>::iterator it = _priorCycleContigs.begin(); it != _priorCycleContigs.end(); ++it){ (*it)->deepDelete(); }
  _priorCycleContigs.clear();

  ostringstream outMessage;
  outMessage << "num contigs leftover: " << leftoverContigs.size();
  _listener->message(outMessage.str());

  // isolate just the read files that are appropriate for this cycle
  set<ParameterizedReadFile*> thisCycleReadFiles;
  for (set<ParameterizedReadFile*>::iterator it = _pairedReadFiles.begin(); it != _pairedReadFiles.end(); ++it){
    if ( (*it)->useThisCycle(cycleNum) ){ thisCycleReadFiles.insert( (*it) ); }
  }

  // this will be augmented later
  //ReadPairFilter* rpf = _rpf->copy();

  // create the extend cycle and copy over the contig sets
  ReadPairFilter* rpfDouble = new ReadPairFilterNull();
  ExtendCycle* ec = new ExtendCycle(cycleNum, &thisCycleReadFiles, _rpf, rpfDouble, &_currentCycleContigs, &leftoverContigs, _listener);
  delete rpfDouble;
  _priorCycleContigs.insert(_currentCycleContigs.begin(), _currentCycleContigs.end());
  _currentCycleContigs.clear();

  // run the cycle and collect, filter, and trim the output
  EcoFilter* sizeFilter = _cycleInfo->getSizeFilter();
  EcoFilter* fullFilter = _cycleInfo->getFullEcoFilter();
  ec->runCycle(_miniAlignParams, _metaAlignParams, _paramsDeBruijn, _cycleInfo->maxContigLinkages(), sizeFilter, fullFilter);
  delete sizeFilter;
  delete fullFilter;
  ec->getContigs(&_currentCycleContigs);
  delete ec;

  _cycleInfo->executeTrim( &_currentCycleContigs, cycleNum );

  // identify those contigs that did not change during this cycle
  _currentCycleLeftovers.clear();
  findLeftoverContigs(&_currentCycleContigs, &_priorCycleContigs, &_currentCycleLeftovers);
}


long Assembler::getTotalSize(){
  long count = 0;
  for (set<ScoredSeq*>::iterator itS = _currentCycleContigs.begin(); itS != _currentCycleContigs.end(); ++itS){
    count += (*itS)->size();
  }
  return count;
}

long Assembler::getMaxSize(){
  long maxSize = 0;
  for (set<ScoredSeq*>::iterator itS = _currentCycleContigs.begin(); itS != _currentCycleContigs.end(); ++itS){
    if ( (*itS)->size() > maxSize ){ maxSize = (*itS)->size(); }
  }
  return maxSize;
}


long Assembler::getN50(){
  vector<long> contigLengths;
  for (set<ScoredSeq*>::iterator seqIt = _currentCycleContigs.begin(); seqIt != _currentCycleContigs.end(); ++seqIt){
    contigLengths.push_back( (*seqIt)->size() );
  }
  sort( contigLengths.begin(), contigLengths.end() );
  long halfSize = getTotalSize() / 2;

  long n50 = 0;
  vector<long>::iterator sizeIt = contigLengths.begin();
  while ( halfSize > 0 and sizeIt != contigLengths.end() ){
    n50 = (*sizeIt); // current n50; will only be updated if half the total size is not reached
    halfSize -= (*sizeIt);
    ++sizeIt;
  }
  return n50;
}
bool reverseN50Helper(long i,long j) { return (i>j); }


#endif



/*
THE AUTHOR: generated by www.degraeve.com/img2txt.php
::::;::::;::::i:,,,tiiiitt;;,;,,,;,,,,,,
::::;::::;:,i:i;i;;jfGLfji:,,;,,,;,,,,,,
::::;::::,:;,jGGfLGLLGLffLtit;,,,;,,,,,,
::::,::::;;jLGGGEEEDGGGLjfLLt;::,;,,,,,,
::::,::::ijLDGDGLKWKEDGGLffLt;,::;:,,,,,
::::,:::ifDGDDGGDEKWKEDLLLGGft:::;::,,,:
::::,::;tiLDDEDDDDDEKKKKEDLLLLi,:;,::,,,
:..:,::jitEDKKDGLLLLGGDEEDDfGLf,:;::,,,,
:::::::L,LDEWDLLffffLLGDDDDfLGLi:,:::,,:
..::::;jiDDKKLLLfffffffLGDEGLLLf:,:.::::
...:,.;:tfEWDGLLffffjjjjjLEDLDLf::...:.:
:...:..;iEEKGGLLfffjfjtttjLDDDLf::...::.
....:t:.LEEKLGLLfffjjjttiitDEEGLL:....:.
:...:LGLEDDDGLLffLLffjttiiijEKGj;:......
:...:jKKKKWELDDDKEDLfjjjjjiiLEEj::......
..:.tiEEKKWDWWKKWKKDffDKDGj;DEDG;:....:.
..:;iGKEKKKLWKEKKWKELEWEEGG#EEDEKtj;,;;,
ii;iiLEEDKDGKDEEEKEDjEEEGfGt#EDDEDi,,,,,
i;;iiiKKKKDLGGGDEEELtWLEEGjtGGGDDL,,,,,,
iiiiiiGfGEDLLLLGDDELiGfLLftfLGEDLfi;;iii
iiiiiiitKDDLLGEEEEGfiiGLLji;LGEGijiiiiii
iiiitiitLfDLGGDDDLGLiitELfi;jDGiijiiiiii
iiiiiiiittGLGDEEEEEDfLfDDLtiGLiiijiiiitt
tttttiiiitEGGGEDDEDDGffLGDtiGLi;itiiiiii
fffffffLGfEDGGDDWEEGGLLKGLjLjfjiitiiiiii
ttttttttitLEGGLGDEGLjDftftiDjiiiitiiiiii
#########WL#GGGGGDDGGjitftDKjiiiitiiiitt
WWWWWWWWKKfWLGGGGGGGLjtttftiiiiiijtttttt
WWWWWDWWWfEKLGGDLGLLfjtittiiitiiijtttttt
WWWWWKKKfGEDGGGDDGLLfttjitiiitiiitiiiitt
WWWWWKLfLfLDGGGDEDGGLjfjjiittttttftttttj
WWWLLGGLLLGEGGGGDDDDDLjtKG;;tttttftttjjj
jjfjjLLfLDKKGGGGGDGGjtttWD;jt;tttftjjjjj
ttjLLLffLLDKDGDGGGGfjtttKf;Lfi,;tjjjjjjj
tjDLDfGLtjGEDGDGDDDfjjjGWL;;tL;,;:tjjjjf
fGLDLLtjjtfDDDDDDDGLjfLEED,iGfLt,,;::;tj
jfLDfjGGGLtLEDGDDGGLLGELfGt;ijGD;ti;;,,j
LjfDfLGKWWWjfDGGDDDDGDjGjft;;fDLij,ti,:D
fDLLLDEKKWKGLjGLGGDGGtLGjfD:itfjfjt,j;tD
LLDGGEDWDEDGDjjtijLitfDjfjKKitDDjff,fiiL
DjLDfDGGGGEGGGjjttjtGfffGDKWKfGDiGL;tiji
jDLjDfGLLLGDGEftiiijtfLLfGEEWttLfLLf;jit
fLEGLjGjffGGGGEftitjfjGjtjjLKiLGttGftjit
GjfDjffjjfLLLLjDfjfffGLti;jjtiGL;fELfijj
LfjGfffjjfLLfffjDLLfGGjji;ttfLLLtEDLLjGj
fLGGffjjjjfGjjjjfGjLGLttiijtttjGjjfDiLLi
LffLjfjjjffGjjjfLGfGDGjtttttjtjjtfLLfff;
LfffjjjjfffGjjffLtfEDDjtfjjtjjffjffjtfjt
GLffjjfjfffGjjfftiGDDGLfffjtjjfLjfGGjjLj
*/
